import functools
from typing import Any, Callable, TypeVar, cast

# Define a generic type variable for the function
F = TypeVar("F", bound=Callable[..., Any])


class RequiresNamedParams(TypeError):
    """
    Custom exception raised when a function is called with positional arguments
    instead of named arguments.
    """

    pass


def require_named_params(
    func: F | None = None, *, example_params: tuple[str, str] | None = None
) -> Callable[[F], F]:
    """
    Decorator that catches TypeError exceptions related to positional arguments
    and re-raises them with a more helpful error message.

    Args:
        func: The function to decorate
        example_params: Optional tuple of (param_name, param_value) to show in the error message
                        Example: example_params=("user_id", "12345")

    Returns:
        The decorated function with improved error messages for positional arguments
    """

    def decorator(func: F) -> F:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            try:
                return func(*args, **kwargs)
            except TypeError as e:
                # Check if this is the "takes X positional arguments but Y were given" error
                if "positional argument" in str(e) and "were given" in str(e):
                    func_name = func.__qualname__

                    # Use custom example if provided, otherwise use generic example
                    if example_params:
                        param_name, param_value = example_params
                        param_example = f"{func_name}({param_name}='{param_value}')"
                        value_example = f"'{param_value}'"
                    else:
                        param_example = f"{func_name}(name='value')"
                        value_example = "'value'"

                    message = (
                        f"{str(e)}. '{func_name}' only accepts named parameters.\n"
                        f"Try using named parameters like: {param_example} instead of {func_name}({value_example})"
                    )
                    raise RequiresNamedParams(message) from None
                # Re-raise other TypeErrors unchanged
                raise

        return cast(F, wrapper)

    # Handle both @require_named_params and @require_named_params(example_params="...")
    if func is None:
        return decorator
    return decorator(func)



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from math import ceil
from pathlib import Path
from typing import Iterator

import pandas as pd
from nptdms import TdmsChannel, TdmsFile, TdmsGroup
from pandas.io.parsers import TextFileReader

from synnax.exceptions import ValidationError
from synnax.io.matcher import new_extension_matcher
from synnax.io.protocol import (
    BaseReader,
    ChannelMeta,
    ColumnFileReader,
    File,
    ReaderType,
)

TDMSMatcher = new_extension_matcher(["tdms"])


class TDMSReader(TDMSMatcher):  # type: ignore
    """A ColReader implementation for TDMS files."""

    channel_keys: set[str]
    chunk_size: int

    _current_chunk: int
    _n_chunks: int | None

    _path: Path
    _channels: list[ChannelMeta] | None

    # Doing a protocol implementation check here because
    # it's hard for pyright to handle factories that return
    # protocol classes.
    def _(self) -> ColumnFileReader:
        return self

    def __init__(
        self,
        path: Path,
        keys: list[str] | None = None,
        chunk_size: int | None = None,
    ):
        self._path = path
        self.chunk_size = chunk_size or int(1e5)

        self._channels = None
        self._current_chunk = 0
        self._n_chunks = None
        if keys is not None:
            self.channel_keys = set(keys)
        else:
            self.set_keys_from_file()

    def channels(self) -> list[ChannelMeta]:
        """:returns : a list of channel metadata for the file."""
        if self._channels is not None:
            return self._channels

        self._channels: list[ChannelMeta] = list()
        with TdmsFile.open(self._path) as tdms_file:
            for group in tdms_file.groups():
                for channel in group.channels():
                    self._channels.append(
                        ChannelMeta(name=channel.name, meta_data=dict())
                    )

        return self._channels

    def set_chunk_size(self, chunk_size: int):
        self.chunk_size = chunk_size

    def set_keys(self, keys: list[str]):
        self.channel_keys = set(keys)

    def set_keys_from_file(self) -> set[str]:
        self.channel_keys = set([ch.name for ch in self.channels()])
        return self.channel_keys

    @classmethod
    def type(cls) -> ReaderType:
        """:returns : the type of reader."""
        return ReaderType.Column

    def path(self) -> Path:
        """:returns: the path to the file."""
        return self._path

    def nsamples(self) -> int:
        """:returns: the number of samples in the file."""
        return self.chunk_size * self.n_chunks * len(self.channels())

    def seek_first(self):
        """Seeks the reader to the first sample in the file."""
        self._current_chunk = 0

    def read(self, *keys: str) -> pd.DataFrame:
        """Reads a dataframe with chunk size * number of columns samples. The returned
        dataframe contains columns for each key in keys.
        """
        # if we already read everything, return an empty chunk
        if self._current_chunk >= self.n_chunks:
            return pd.DataFrame()

        # if keys is empty, use default keys
        keys: set[str] = self.channel_keys if (len(keys) == 0) else set(keys)

        # https://nptdms.readthedocs.io/en/stable/reading.html
        # https://github.com/adamreeve/npTDMS/issues/263
        # https://nptdms.readthedocs.io/en/stable/reading.html
        data = dict()
        with TdmsFile.open(self._path) as tdms_file:
            for group in tdms_file.groups():
                for channel in group.channels():
                    if channel.name in keys:
                        data[channel.name] = channel[
                            self._current_chunk
                            * self.chunk_size : (self._current_chunk + 1)
                            * self.chunk_size
                        ]
            self._current_chunk += 1

        return pd.DataFrame(data)

    def __iter__(self) -> Iterator[pd.DataFrame]:
        return (self.read() for _ in range(self.n_chunks))

    @property
    def n_chunks(self) -> int:
        """Returns number of chunks in a channel.
        Assumes every channel has the same number of chunks
        """
        if self._n_chunks is None:
            # assume every channel has the same number of data
            with TdmsFile.open(self._path) as tdms_file:
                group0 = tdms_file.groups()[0]
                channel0 = group0.channels()[0]
                self._n_chunks = ceil(len(channel0) / self.chunk_size)

        assert self._n_chunks is not None
        return self._n_chunks



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from alamos import Instrumentation
from freighter import (
    URL,
    AsyncMiddleware,
    AsyncStreamClient,
    AsyncWebsocketClient,
    HTTPClient,
    JSONCodec,
    Middleware,
    MsgPackCodec,
    StreamClient,
    UnaryClient,
    WebsocketClient,
    async_instrumentation_middleware,
    instrumentation_middleware,
)
from urllib3 import Retry, Timeout

from synnax.telem import Size, TimeSpan


class Transport:
    url: URL
    stream: StreamClient
    stream_async: AsyncStreamClient
    unary: UnaryClient
    secure: bool

    def __init__(
        self,
        url: URL,
        instrumentation: Instrumentation,
        secure: bool = False,
        open_timeout: TimeSpan = TimeSpan.SECOND * 5,
        read_timeout: TimeSpan = TimeSpan.SECOND * 5,
        keep_alive: TimeSpan = TimeSpan.SECOND * 30,
        max_retries: int = 3,
    ) -> None:
        self.url = url.child("/api/v1/")
        ws_args = {
            "base_url": self.url,
            "encoder": MsgPackCodec(),
            "max_message_size": int(Size.MB * 5),
            "secure": secure,
            "open_timeout": open_timeout.seconds,
            "close_timeout": read_timeout.seconds,
        }
        self.stream = WebsocketClient(**ws_args)
        # We need to update these here because the websocket client doesn't support
        # the same arguments as the async websocket client.
        ws_args["ping_interval"] = keep_alive.seconds
        ws_args["ping_timeout"] = 180
        self.stream_async = AsyncWebsocketClient(**ws_args)
        self.unary = HTTPClient(
            url=self.url,
            codec=JSONCodec(),
            secure=secure,
            timeout=Timeout(connect=open_timeout.seconds, read=read_timeout.seconds),
            retries=Retry(total=max_retries),
        )
        self.use(instrumentation_middleware(instrumentation))
        self.use_async(async_instrumentation_middleware(instrumentation))

    def use(self, *middleware: Middleware):
        self.unary.use(*middleware)
        self.stream.use(*middleware)

    def use_async(self, *middleware: AsyncMiddleware):
        self.stream_async.use(*middleware)



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import json
from typing import Any

from pydantic import BaseModel

from synnax.channel import ChannelKey
from synnax.hardware.task import JSONConfigMixin, MetaTask, StarterStopperMixin, Task
from synnax.telem import CrudeRate, Rate

TYPE = "sequence"


class Config(BaseModel):
    """Configuration for a sequence task."""

    rate: Rate
    read: list[ChannelKey]
    write: list[ChannelKey]
    script: str
    globals: dict[str, Any] = {}


ZERO_CONFIG = Config(rate=Rate(10), read=[], write=[], script="", globals={})


class StateDetails(BaseModel):
    """State details for a sequence task."""

    running: bool
    message: str


class Sequence(StarterStopperMixin, JSONConfigMixin, MetaTask):
    """A task for executing control sequences on a Synnax cluster."""

    TYPE = TYPE
    config: Config
    _internal: Task

    def __init__(
        self,
        internal: Task | None = None,
        *,
        name: str = "Control Sequence",
        rate: CrudeRate = 10,
        read: list[ChannelKey] = [],
        write: list[ChannelKey] = [],
        script: str = "",
        globals: dict[str, Any] = {},
    ):
        if internal is not None:
            self._internal = internal
            self.config = Config.model_validate_json(internal.config)
            return
        self._internal = Task(name=name, type=self.TYPE)
        self.config = Config(
            rate=Rate(rate), read=read, write=write, script=script, globals=globals
        )



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.


from freighter import (
    AsyncMiddleware,
    AsyncNext,
    Context,
    Middleware,
    Next,
    Payload,
    UnaryClient,
)

from synnax.exceptions import ExpiredToken, InvalidToken
from synnax.user.payload import User
from synnax.util.send_required import send_required


class InsecureCredentials(Payload):
    username: str
    password: str


class TokenResponse(Payload):
    token: str
    user: User


AUTHORIZATION_HEADER = "Authorization"
TOKEN_REFRESH_HEADER = "Refresh-Token"
RETRY_ON_ERRORS = (InvalidToken, ExpiredToken)
TOKEN_PREFIX = "Bearer "


class AuthenticationClient:
    _LOGIN_ENDPOINT = "/auth/login"

    client: UnaryClient
    username: str
    password: str
    token: str
    user: User
    authenticated: bool

    def __init__(
        self,
        transport: UnaryClient,
        username: str,
        password: str,
    ) -> None:
        self.client = transport
        self.username = username
        self.password = password
        self.authenticated = False

    def authenticate(self) -> None:
        res = send_required(
            self.client,
            self._LOGIN_ENDPOINT,
            InsecureCredentials(username=self.username, password=self.password),
            TokenResponse,
        )
        self.token = res.token
        self.user = res.user
        self.authenticated = True

    def middleware(self) -> list[Middleware]:
        def mw(ctx: Context, _next: Next):
            if not self.authenticated:
                self.authenticate()

            ctx.set(AUTHORIZATION_HEADER, TOKEN_PREFIX + self.token)
            out_ctx, exc = _next(ctx)

            if isinstance(exc, RETRY_ON_ERRORS):
                self.authenticated = False
                out_ctx, exc = mw(ctx, _next)

            self.maybe_refresh_token(out_ctx)
            return out_ctx, exc

        return mw

    def async_middleware(self) -> list[AsyncMiddleware]:
        async def mw(ctx: Context, _next: AsyncNext):
            if not self.authenticated:
                self.authenticate()

            ctx.set(AUTHORIZATION_HEADER, TOKEN_PREFIX + self.token)
            out_ctx, exc = await _next(ctx)

            if isinstance(exc, RETRY_ON_ERRORS):
                self.authenticated = False
                out_ctx, exc = await mw(ctx, _next)

            self.maybe_refresh_token(out_ctx)
            return out_ctx, exc

        return mw

    def maybe_refresh_token(
        self,
        ctx: Context,
    ) -> None:
        refresh = ctx.get(TOKEN_REFRESH_HEADER, None)
        if refresh is not None:
            self.token = refresh



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import json

from freighter import Payload

from synnax.status import INFO_VARIANT, Variant


class TaskPayload(Payload):
    key: int = 0
    name: str = ""
    type: str = ""
    config: str = ""
    snapshot: bool = False


class TaskState(Payload):
    task: int = 0
    variant: Variant = INFO_VARIANT
    key: str = ""
    details: dict

    def __init__(
        self,
        task: int = 0,
        variant: Variant = INFO_VARIANT,
        key: str = "",
        details: dict | str = "",
    ):
        if isinstance(details, str):
            details = json.loads(details)
        super().__init__(task=task, variant=variant, key=key, details=details)



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from uuid import UUID

from freighter import Payload

from synnax.telem import TimeRange


class RangePayload(Payload):
    """Network transportable payload representing a range."""

    key: UUID = UUID(int=0)
    name: str = ""
    time_range: TimeRange
    color: str = "#000000"


RangeKey = UUID | str
"""The type for the key of a Range. A UUID."""
RangeName = str
"""The type for the name of a Range. A string."""
RangeKeys = tuple[UUID] | list[UUID]
"""The type for the keys of a Range. A tuple or list of UUIDs."""
RangeNames = tuple[str] | list[str]
"""The type for the names of a Range. A tuple or list of strings."""
RangeParams = RangeKeys | RangeNames | RangeKey | RangeName
"""Parameters that can be used to query a range"""



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import overload

from alamos import NOOP, Instrumentation
from freighter import Empty, Payload, UnaryClient, send_required

from synnax.hardware.rack.payload import Rack
from synnax.util.normalize import check_for_none, override


class _CreateRequest(Payload):
    racks: list[Rack]


class _CreateResponse(Payload):
    racks: list[Rack]


class _DeleteRequest(Payload):
    keys: list[int]


class _RetrieveRequest(Payload):
    keys: list[int] | None = None
    names: list[str] | None = None
    search: str | None = None
    embedded: bool | None = None
    host_is_node: bool | None = None
    offset: int | None = None
    limit: int | None = None


class _RetrieveResponse(Payload):
    racks: list[Rack] | None = None


_CREATE_ENDPOINT = "/hardware/rack/create"
_DELETE_ENDPOINT = "/hardware/rack/delete"
_RETRIEVE_ENDPOINT = "/hardware/rack/retrieve"


class Client:
    _client: UnaryClient
    _embedded_rack: Rack | None = None
    instrumentation: Instrumentation

    def __init__(
        self,
        client: UnaryClient,
        instrumentation: Instrumentation = NOOP,
    ) -> None:
        self._client = client
        self.instrumentation = instrumentation

    @overload
    def create(self, *, key: int = 0, name: str = "") -> Rack: ...

    @overload
    def create(self, rack: Rack) -> Rack: ...

    @overload
    def create(self, racks: list[Rack]) -> list[Rack]: ...

    def create(
        self,
        racks: Rack | list[Rack] | None = None,
        *,
        key: int = 0,
        name: str = "",
    ) -> Rack | list[Rack]:
        is_single = True
        if racks is None:
            racks = [Rack(key=key, name=name)]
        elif isinstance(racks, Rack):
            racks = [racks]
        else:
            is_single = False
        req = _CreateRequest(racks=racks)
        res = send_required(self._client, _CREATE_ENDPOINT, req, _CreateResponse)
        if is_single:
            return res.racks[0]
        return res.racks

    def delete(self, keys: list[int]):
        req = _DeleteRequest(keys=keys)
        send_required(self._client, _DELETE_ENDPOINT, req, Empty)

    @overload
    def retrieve(
        self,
        key: int | None = None,
        name: str | None = None,
        embedded: bool = False,
        host_is_node: bool = False,
    ) -> Rack: ...

    def retrieve(
        self,
        key: int | None = None,
        name: str | None = None,
        keys: list[int] | None = None,
        names: list[str] | None = None,
        *,
        host_is_node: bool = False,
        embedded: bool = False,
    ) -> list[Rack]:
        is_single = check_for_none(keys, names)
        res = send_required(
            self._client,
            _RETRIEVE_ENDPOINT,
            _RetrieveRequest(
                keys=override(key, keys),
                names=override(name, names),
                host_is_node=host_is_node,
                embedded=embedded,
            ),
            _RetrieveResponse,
        )
        return res.racks[0] if is_single else res.racks

    def retrieve_embedded_rack(self) -> Rack:
        if self._embedded_rack is None:
            self._embedded_rack = self.retrieve(embedded=True, host_is_node=True)
        return self._embedded_rack



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import Any

# We dynamically define all of these operators to allow for interoperability with numpy
# arrays.
_NP_COMPARISON_OPERATORS = [
    "__eq__",
    "__ne__",
    "__lt__",
    "__gt__",
    "__le__",
    "__ge__",
    "__add__",
    "__sub__",
    "__mul__",
    "__floordiv__",
    "__mod__",
    "__pow__",
    "__lshift__",
    "__rshift__",
    "__and__",
    "__xor__",
    "__or__",
    "__neg__",
    "__pos__",
    "__abs__",
    "__invert__",
    "__round__",
    "__floor__",
    "__ceil__",
    "__trunc__",
    "__int__",
    "__float__",
    "__index__",
    "__truediv__",
    "__divmod__",
    "__radd__",
    "__rsub__",
    "__rmul__",
    "__rfloordiv__",
    "__rmod__",
    "__rpow__",
    "__rlshift__",
    "__rrshift__",
    "__rand__",
    "__rxor__",
    "__ror__",
    "__iadd__",
    "__isub__",
    "__imul__",
    "__ifloordiv__",
    "__imod__",
    "__ipow__",
    "__ilshift__",
    "__irshift__",
    "__iand__",
    "__ixor__",
    "__ior__",
    "__rtruediv__",
    "__rdivmod__",
    "__itruediv__",
    "__ifloordiv__",
    "__matmul__",
    "__rmatmul__",
    "__imatmul__",
]


def overload_comparison_operators(cls: type[Any], method: str) -> type[Any]:
    return forward_methods(cls, method, _NP_COMPARISON_OPERATORS)


def forward_methods(cls: type[Any], to: str, methods: list[str]) -> type[Any]:
    for method in methods:
        setattr(
            cls,
            method,
            lambda self, *args, method=method: getattr(getattr(self, to)(), method)(
                *args
            ),
        )
    return cls



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import overload

from alamos import NOOP, Instrumentation, trace
from freighter import Empty, Payload, UnaryClient, send_required

from synnax.exceptions import NotFoundError
from synnax.hardware.device.payload import Device
from synnax.util.normalize import check_for_none, normalize, override


class _CreateRequest(Payload):
    devices: list[Device]


_CreateResponse = _CreateRequest


class _DeleteRequest(Payload):
    keys: list[str]


class _RetrieveRequest(Payload):
    keys: list[str] | None = None
    names: list[str] | None = None
    makes: list[str] | None = None
    models: list[str] | None = None
    locations: list[str] | None = None
    ignore_not_found: bool = False


class _RetrieveResponse(Payload):
    devices: list[Device] | None = None


_CREATE_ENDPOINT = "/hardware/device/create"
_DELETE_ENDPOINT = "/hardware/device/delete"
_RETRIEVE_ENDPOINT = "/hardware/device/retrieve"


class Client:
    _client: UnaryClient
    instrumentation: Instrumentation = NOOP

    def __init__(
        self,
        client: UnaryClient,
        instrumentation: Instrumentation = NOOP,
    ) -> None:
        self._client = client
        self.instrumentation = instrumentation

    @overload
    def create(
        self,
        *,
        key: str = "",
        location: str = "",
        rack: int = 0,
        name: str = "",
        make: str = "",
        model: str = "",
        properties: str = "",
    ): ...

    @overload
    def create(self, devices: Device): ...

    @overload
    def create(self, devices: list[Device]): ...

    def create(
        self,
        devices: list[Device] | Device | None = None,
        *,
        key: str = "",
        location: str = "",
        rack: int = 0,
        name: str = "",
        make: str = "",
        model: str = "",
        properties: str = "",
    ):
        is_single = not isinstance(devices, list)
        if devices is None:
            devices = [
                Device(
                    key=key,
                    location=location,
                    rack=rack,
                    name=name,
                    make=make,
                    model=model,
                    properties=properties,
                )
            ]
        req = _CreateRequest(devices=normalize(devices))
        res = send_required(
            self._client,
            _CREATE_ENDPOINT,
            req,
            _CreateResponse,
        )
        return res.devices[0] if is_single else res.devices

    def delete(self, keys: list[str]) -> None:
        req = _DeleteRequest(keys=keys)
        send_required(self._client, _DELETE_ENDPOINT, req, Empty)

    @overload
    def retrieve(
        self,
        *,
        key: str | None = None,
        make: str | None = None,
        name: str | None = None,
        model: str | None = None,
        location: str | None = None,
    ) -> Device: ...

    @overload
    def retrieve(
        self,
        *,
        keys: list[str] | None = None,
        makes: list[str] | None = None,
        models: list[str] | None = None,
        names: list[str] | None = None,
        locations: list[str] | None = None,
        ignore_not_found: bool = False,
    ) -> list[Device]: ...

    @trace("debug")
    def retrieve(
        self,
        *,
        key: str | None = None,
        make: str | None = None,
        model: str | None = None,
        name: str | None = None,
        location: str | None = None,
        keys: list[str] | None = None,
        makes: list[str] | None = None,
        models: list[str] | None = None,
        names: list[str] | None = None,
        locations: list[str] | None = None,
        ignore_not_found: bool = False,
    ) -> list[Device] | Device:
        is_single = check_for_none(keys, makes, models, locations, names)
        res = send_required(
            self._client,
            _RETRIEVE_ENDPOINT,
            _RetrieveRequest(
                keys=override(key, keys),
                makes=override(make, makes),
                models=override(model, models),
                locations=override(location, locations),
                names=override(name, names),
                ignore_not_found=ignore_not_found,
            ),
            _RetrieveResponse,
        )
        if is_single:
            if len(res.devices) > 0:
                return res.devices[0]
            if ignore_not_found:
                return None
            raise NotFoundError("Device not found")
        return res.devices



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import Literal

SUCCESS_VARIANT = "success"
INFO_VARIANT = "info"
WARNING_VARIANT = "warning"
ERROR_VARIANT = "error"
DISABLED_VARIANT = "disabled"
LOADING_VARIANT = "loading"

Variant = Literal[
    "success",
    "info",
    "warning",
    "error",
    "disabled",
    "loading",
]
"""Represents the variant of a status message."""



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import uuid

from alamos import NOOP, Instrumentation, trace
from freighter import Payload, UnaryClient

from synnax.ranger.payload import RangeKey, RangeName, RangePayload
from synnax.util.normalize import normalize


class _Request(Payload):
    keys: list[uuid.UUID | str] | None = None
    names: list[str] | None = None
    term: str | None = None


class _Response(Payload):
    ranges: list[RangePayload] | None


class RangeRetriever:
    __ENDPOINT = "/range/retrieve"
    __client: UnaryClient
    instrumentation: Instrumentation = NOOP

    def __init__(
        self,
        client: UnaryClient,
        instrumentation: Instrumentation = NOOP,
    ) -> None:
        self.__client = client
        self.instrumentation = instrumentation

    @trace("debug")
    def retrieve(
        self,
        key: RangeKey | None = None,
        name: RangeName | None = None,
        names: list[RangeName] | None = None,
        keys: list[RangeKey] | None = None,
    ) -> list[RangePayload]:
        if key is not None:
            keys = normalize(key)
        if name is not None:
            names = normalize(name)
        return self.__execute(_Request(keys=keys, names=names))

    @trace("debug")
    def search(self, term: str) -> list[RangePayload]:
        return self.__execute(_Request(term=term))

    def __execute(self, req: _Request) -> list[RangePayload]:
        res, exc = self.__client.send(self.__ENDPOINT, req, _Response)
        if exc is not None:
            raise exc
        if res.ranges is None:
            return list()
        return res.ranges



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

import click
import matplotlib.pyplot as plt
import numpy as np

import synnax as sy
from synnax.cli import default
from synnax.cli.channel import select_channel
from synnax.cli.connect import connect_client
from synnax.cli.warning import warning


@click.command()
@click.pass_context
def check_timing(ctx: click.Context) -> None:
    """Check the timing characteristics of a time channel."""
    warning(ctx)
    client = connect_client(default.context())
    if client is None:
        return

    # Get all channels and filter for time channels
    channels = client.channels.retrieve(["*"])
    time_channels = [ch for ch in channels if ch.data_type == sy.DataType.TIMESTAMP]

    if not time_channels:
        ctx.console.error("No time channels found in the database")
        return

    # Let user select the time channel
    time_channel = select_channel(default.context(), time_channels, key="name")
    if time_channel is None:
        return

    # Let user specify duration
    duration = default.context().console.ask(
        "How long would you like to collect samples for (in seconds)?",
        type_=int,
        default=10,
    )
    span = sy.TimeSpan.SECOND * duration

    # Collect samples
    offsets, diffs, times, local_start, local_end = collect_samples(
        client, time_channel.key, span
    )

    # Replace the individual report calls with the combined report
    create_timing_report(times, offsets, diffs, local_start, local_end)


def collect_samples(
    client: sy.Synnax,
    time_channel: sy.channel.ChannelKey,
    span: sy.TimeSpan,
):
    # Tracks the offset between the local clock and the time channel
    offsets: list[sy.TimeSpan] = list()
    # Tracks the spacing between the samples inside individual reads
    diffs: list[sy.TimeSpan] = list()
    # Tracks all the times received
    times: list[sy.TimeStamp] = list()
    # Track local start time
    local_start = sy.TimeStamp.now()

    with client.open_streamer(time_channel) as streamer:
        now = local_start
        end = now + span
        while now < end:
            now = sy.TimeStamp.now()
            data = streamer.read()[time_channel]
            offset = sy.TimeSpan(sy.TimeStamp.now() - sy.TimeStamp(data[-1]))
            offsets.append(offset)
            diff = sy.TimeSpan(data[-1] - data[-2])
            diffs.append(diff)
            times.extend(data)

    local_end = sy.TimeStamp.now()
    return offsets, diffs, times, local_start, local_end


def create_timing_report(
    times: list[sy.TimeStamp],
    offsets: list[sy.TimeSpan],
    diffs: list[sy.TimeSpan],
    local_start: sy.TimeStamp,
    local_end: sy.TimeStamp,
) -> None:
    # Set dark theme
    plt.style.use("dark_background")

    # Create figure with subplots
    fig = plt.figure(figsize=(15, 10))
    fig.suptitle("Timing Analysis Report", fontsize=16)

    # Time Values Derivative vs Index plot
    ax1 = plt.subplot(2, 2, 1)
    indices = np.arange(len(times) - 1)
    # Convert timestamps to nanoseconds and then calculate differences
    derivatives = np.diff(times)

    # Check for negative derivatives
    negative_indices = np.where(derivatives < 0)[0]
    if len(negative_indices) > 0:
        print("\nWARNING: Negative derivatives detected!")
        print("Negative derivatives found at indices:", negative_indices)
        print("Derivative values at these indices:", derivatives[negative_indices])
        print("Original timestamps at these locations:")
        for idx in negative_indices:
            print(f"Index {idx}: {times[idx]} -> {times[idx+1]}")

    ax1.plot(indices, derivatives, "cyan", marker="o", markersize=2)
    ax1.set_title("Time Value Derivatives vs Index")
    ax1.set_xlabel("Index")
    ax1.set_ylabel("Time Derivative (ns)")
    ax1.grid(True, alpha=0.2)

    # Clock Offsets plot
    ax2 = plt.subplot(2, 2, 2)
    offsets_array = np.array([float(o.microseconds) for o in offsets])
    offset_mean = np.mean(offsets_array)
    offset_std = np.std(offsets_array)

    # Find offset outliers
    offset_outliers = offsets_array[
        np.abs(offsets_array - offset_mean) > 5 * offset_std
    ]

    bins_offset = np.concatenate(
        [np.linspace(min(offsets_array), max(offsets_array), 1000)]
    )
    hist_offset, bins_offset, _ = ax2.hist(
        offsets_array, bins=bins_offset, alpha=0.7, color="cyan"
    )

    x_offset = np.linspace(min(offsets_array), max(offsets_array), 100)
    gaussian_offset = hist_offset.max() * np.exp(
        -((x_offset - offset_mean) ** 2) / (2 * offset_std**2)
    )
    ax2.plot(x_offset, gaussian_offset, "magenta", lw=2, label="Gaussian fit")

    ax2.set_title(
        "Distribution of Clock Offsets\nDifference between time channel and local machine reception time"
    )
    ax2.set_xlabel("Offset (microseconds)")
    ax2.set_ylabel("Count")
    ax2.legend()
    ax2.grid(True, alpha=0.2)

    # Time Differences plot
    ax3 = plt.subplot(2, 2, 3)
    diffs_array = np.array([float(d.microseconds) for d in diffs])
    diff_mean = np.mean(diffs_array)
    diff_std = np.std(diffs_array)

    # Find diff outliers
    diff_outliers = diffs_array[np.abs(diffs_array - diff_mean) > 5 * diff_std]

    # Convert time differences to rates
    rates = [sy.Rate(d) for d in diffs]
    avg_rate = np.mean([float(r) for r in rates])

    bins_diff = np.concatenate([np.linspace(min(diffs_array), max(diffs_array), 500)])
    hist_diff, bins_diff, _ = ax3.hist(
        diffs_array, bins=bins_diff, alpha=0.7, color="cyan"
    )

    x_diff = np.linspace(min(diffs_array), max(diffs_array), 100)
    gaussian_diff = hist_diff.max() * np.exp(
        -((x_diff - diff_mean) ** 2) / (2 * diff_std**2)
    )
    ax3.plot(x_diff, gaussian_diff, "magenta", lw=2, label="Gaussian fit")

    ax3.set_title("Distribution of Time Differences")
    ax3.set_xlabel("Time Difference (microseconds)")
    ax3.set_ylabel("Count")
    ax3.legend()
    ax3.grid(True, alpha=0.2)

    # Calculate stream rates and times
    stream_start = times[0]
    stream_end = times[-1]
    stream_duration = sy.TimeSpan(stream_end - stream_start)
    stream_rate = sy.Rate(stream_duration / len(times))
    local_duration = sy.TimeSpan(local_end - local_start)

    # Statistics text box
    ax4 = plt.subplot(2, 2, 4)
    ax4.axis("off")
    stats_text = (
        f"Statistics:\n\n"
        f"Clock Offsets:\n"
        f"Mean: {offset_mean:.2f} µs\n"
        f"Std Dev: {offset_std:.2f} µs\n"
        f"Outliers (>5σ): {len(offset_outliers)} values\n\n"
        f"Time Differences:\n"
        f"Mean: {diff_mean:.2f} µs\n"
        f"Std Dev: {diff_std:.2f} µs\n"
        f"Outliers (>5σ): {len(diff_outliers)} values\n"
        f"Calculated Rate: {avg_rate:.2f} Hz\n\n"
        f"Stream Statistics:\n"
        f"Local Start: {local_start}\n"
        f"Local End: {local_end}\n"
        f"Local Duration: {local_duration}\n"
        f"Received Start: {sy.TimeStamp(stream_start)}\n"
        f"Received End: {sy.TimeStamp(stream_end)}\n"
        f"Received Duration: {stream_duration}\n"
        f"Average Rate: {stream_rate}"
    )
    ax4.text(
        0.1, 0.5, stats_text, fontsize=12, verticalalignment="center", color="white"
    )

    # Adjust layout and save
    plt.tight_layout()
    plt.savefig("timing_report.png")
    plt.show()
    plt.close()

    # Print statistics to console as well
    print("\nTiming Analysis Results:")
    print(f"Mean offset: {offset_mean:.2f} microseconds")
    print(f"Standard deviation of offsets: {offset_std:.2f} microseconds")
    print(f"Mean time difference: {diff_mean:.2f} microseconds")
    print(f"Standard deviation of time differences: {diff_std:.2f} microseconds")



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import overload

from freighter import (
    EOF,
    AsyncStream,
    AsyncStreamClient,
    ExceptionPayload,
    Payload,
    Stream,
    WebsocketClient,
)
from freighter.websocket import Message

from synnax.channel.payload import ChannelKeys, ChannelParams
from synnax.exceptions import UnexpectedError
from synnax.framer.adapter import ReadFrameAdapter
from synnax.framer.codec import LOW_PERF_SPECIAL_CHAR, WSFramerCodec
from synnax.framer.frame import Frame, FramePayload
from synnax.telem import TimeSpan


class _Request(Payload):
    keys: ChannelKeys
    downsample_factor: int


class _Response(Payload):
    frame: FramePayload


class WSStreamerCodec(WSFramerCodec):
    def encode(self, pld: Message) -> bytes:
        return self.lower_perf_codec.encode(pld)

    def decode(self, data: bytes, pld_t: Message[_Response]) -> object:
        if data[0] == LOW_PERF_SPECIAL_CHAR:
            msg = self.lower_perf_codec.decode(data[1:], pld_t)
            return msg
        frame = self.codec.decode(data, 1)
        return Message(type="data", payload=_Response(frame=frame))


_ENDPOINT = "/frame/stream"


class Streamer:
    """A streamer is used to stream frames of telemetry in real-time from a Synnax
    cluster. It should not be constructed directly, and should instead be created using
    the client's `open_streamer` method.

    To open a streamer, use the `open_streamer` method on the client and pass in the
    list of channels you'd like to stream. Once a new streamer has been opened, you
    can call the `read` method to read the next frame of telemetry. Once done, call
    the `close` method to close the streamer and free all necessary resources. We
    recommend using the streamer as a context manager to ensure that it is closed
    properly.

    Streamers also support the iterator protocol, allowing you to iterate over the
    frames of telemetry as they are received. This is useful when you want to process
    each frame as it is received.

    For detailed documentation, see https://docs.synnaxlabs.com/reference/python-client/stream-data
    """

    _stream: Stream[_Request, _Response]
    _adapter: ReadFrameAdapter

    def __init__(
        self,
        client: WebsocketClient,
        adapter: ReadFrameAdapter,
        downsample_factor: int = 1,
        use_experimental_codec: bool = True,
    ) -> None:
        self._adapter = adapter
        if use_experimental_codec:
            client = client.with_codec(WSStreamerCodec(self._adapter.codec))

        self._stream = client.stream(_ENDPOINT, _Request, _Response)
        self._downsample_factor = downsample_factor
        self._stream.send(
            _Request(keys=self._adapter.keys, downsample_factor=self._downsample_factor)
        )
        _, exc = self._stream.receive()
        if exc is not None:
            raise exc

    @overload
    def read(self, timeout: float | int | TimeSpan) -> Frame | None:
        """Reads the next frame of telemetry from the streamer with a timeout. If no
        frame is received within the timeout, this method will return None. If an error
        occurs while reading the frame, an exception will be raised.

        :param timeout: The maximum amount of time to wait for a frame to be received
        before returning None. This can be a float or integer representing the number
        of seconds, or a synnax TimeSpan object.
        :return: The next frame of telemetry, or None if no frame is received within the
        timeout.
        """
        ...

    @overload
    def read(self) -> Frame:
        """Reads the next frame of telemetry from the streamer, blocking until a frame
        is received. If an error occurs while reading the frame, an exception will be
        raised.
        """
        ...

    def read(self, timeout: float | None = None) -> Frame | None:
        """Reads the next frame of telemetry from the streamer. If a timeout is provided,
        this method will wait for the specified amount of time for a frame to be received.
        If no frame is received within the timeout, this method will return None. If no
        timeout is provided, this method will block until a frame is received.

        If an error occurs while reading the frame, an exception will be raised.

        :param timeout: The maximum amount of time to wait for a frame to be received
        before returning None. If no timeout is provided, this method will block until a
        frame is received.
        :return: The next frame of telemetry, or None if no frame is received within the
        timeout.
        """
        try:
            res, exc = self._stream.receive(TimeSpan.to_seconds(timeout))
            if exc is not None:
                raise exc
            return self._adapter.adapt(Frame(res.frame))
        except TimeoutError:
            return None

    def update_channels(self, channels: ChannelParams):
        """Updates the list of channels to stream. This method will replace the current
        list of channels with the new list, not add to it.

        :param channels: The list of channels to stream.
        :raises NotFoundError: If any of the channels in the list are not found.
        """
        self._adapter.update(channels)
        self._stream.send(
            _Request(keys=self._adapter.keys, downsample_factor=self._downsample_factor)
        )

    def close(self, timeout: float | int | TimeSpan | None = None):
        """Closes the streamer and frees all network resources.

        :param timeout: The maximum amount of time to wait for the server to acknowledge
        the closure before raising a TimeoutError. This can be a float or integer
        representing the number of seconds, or a synnax TimeSpan object. If no timeout
        is provided, this method will block until the server acknowledges the closure.
        """
        exc = self._stream.close_send()
        if exc is not None:
            raise exc
        while True:
            r, exc = self._stream.receive(TimeSpan.to_seconds(timeout))
            if r is not None:
                continue
            if exc is None:
                raise UnexpectedError(
                    f"""Unexpected missing close acknowledgement from server.
                    Please report this issue to the Synnax team.
                    Response: {r}
                    """
                )
            elif not isinstance(exc, EOF):
                raise exc
            break

    def __iter__(self):
        """Returns an iterator object that can be used to iterate over the frames of
        telemetry as they are received. This is useful when you want to process each
        frame as it is received.
        """
        return self

    def __enter__(self):
        """Returns the streamer object when used as a context manager."""
        return self

    def __next__(self):
        """Reads the next frame of telemetry from the streamer."""
        return self.read()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class AsyncStreamer:
    """An asynchronous version of the Streamer class. This class is used to stream
    frames of telemetry in real-time from a Synnax cluster. It should not be constructed
    directly, and should instead be created using the client's `open_streamer` method.

    To open an async streamer, use the `open_async_streamer` method on the client and pass
    in the list of channels you'd like to stream. Once a new async streamer has been opened,
    you can call the `read` method to read the next frame of telemetry. Once done, call the
    `close` method to close the streamer and free all necessary resources. We recommend
    using the async streamer as an async context manager to ensure that it is closed properly.

    Async streamers also support the async iterator protocol, allowing you to iterate over
    the frames of telemetry as they are received. This is useful when you want to process
    each frame as it is received.
    """

    _stream: AsyncStream[_Request, _Response]
    _client: AsyncStreamClient
    _adapter: ReadFrameAdapter

    def __init__(
        self,
        client: AsyncStreamClient,
        adapter: ReadFrameAdapter,
        downsample_factor: int,
    ) -> None:
        self._client = client
        self._adapter = adapter
        self._downsample_factor = downsample_factor

    async def _open(self):
        self._stream = await self._client.stream(_ENDPOINT, _Request, _Response)
        await self._stream.send(
            _Request(
                keys=self._adapter.keys,
                downsample_factor=self._downsample_factor,
            )
        )
        _, exc = await self._stream.receive()
        if exc is not None:
            raise exc

    @property
    def received(self) -> bool:
        """Returns True if a frame has been received, False otherwise."""
        return self._stream.received()

    async def read(self) -> Frame:
        """Reads the next frame of telemetry from the streamer. If an error occurs while
        reading the frame, an exception will be raised.
        """
        res, exc = await self._stream.receive()
        if exc is not None:
            raise exc
        return self._adapter.adapt(Frame(res.frame))

    async def close_loop(self):
        """Closes the sending end of the streamer, requiring the caller to process all
        remaining frames and close acknowledgements by calling read. This method is
        useful for managing the lifecycle of a streamer within a separate event loop or
        thread.
        """
        await self._stream.close_send()

    async def close(self):
        """Close the streamer and free all network resources, waiting for the server to
        acknowledge the close request.
        """
        exc = await self._stream.close_send()
        if exc is not None:
            raise exc
        _, exc = await self._stream.receive()
        if exc is None:
            raise UnexpectedError(
                """Unexpected missing close acknowledgement from server.
                Please report this issue to the Synnax team."""
            )
        elif not isinstance(exc, EOF):
            raise exc

    async def __aenter__(self):
        """Returns the async streamer object when used as an async context manager."""
        return self

    def __aiter__(self):
        """Returns an async iterator object that can be used to iterate over the frames
        of telemetry as they are received. This is useful when you want to process each
        frame as it is received.
        """
        return self

    async def __anext__(self):
        """Reads the next frame of telemetry from the streamer."""
        try:
            return await self.read()
        except EOF:
            raise StopAsyncIteration

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Closes the streamer when used as an async context manager"""
        await self.close()



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

from freighter import Payload


class Authority(int):
    ABSOLUTE: Authority
    """Absolute authority. No other subject can take control of this entity while it is
    active."""
    DEFAULT: Authority


Authority.ABSOLUTE = Authority(255)
Authority.DEFAULT = Authority(1)

CrudeAuthority = int | Authority


class Subject(Payload):
    name: str
    key: str



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from enum import Enum
from typing import Literal, TypeAlias, overload
from uuid import uuid4

from freighter import (
    EOF,
    ExceptionPayload,
    Payload,
    Stream,
    WebsocketClient,
    decode_exception,
)
from freighter.websocket import Message

from synnax.channel.payload import (
    ChannelKey,
    ChannelKeys,
    ChannelName,
    ChannelNames,
    ChannelPayload,
)
from synnax.framer.adapter import WriteFrameAdapter
from synnax.framer.codec import (
    HIGH_PERF_SPECIAL_CHAR,
    LOW_PERF_SPECIAL_CHAR,
    WSFramerCodec,
)
from synnax.framer.frame import CrudeFrame, FramePayload
from synnax.telem import CrudeSeries, CrudeTimeStamp, TimeSpan, TimeStamp
from synnax.telem.control import Authority, CrudeAuthority, Subject
from synnax.util.normalize import normalize


class WriterCommand(int, Enum):
    OPEN = 0
    WRITE = 1
    COMMIT = 2
    SET_AUTHORITY = 3


class WriterMode(int, Enum):
    PERSIST_STREAM = 1
    PERSIST = 2
    STREAM = 3


CrudeWriterMode: TypeAlias = (
    WriterMode
    | Literal["persist_stream"]
    | Literal["persist"]
    | Literal["stream"]
    | int
)


class WriterConfig(Payload):
    authorities: list[int] = Authority.ABSOLUTE
    control_subject: Subject = Subject(name="", key=str(uuid4()))
    start: TimeStamp | None = None
    keys: ChannelKeys
    mode: WriterMode = WriterMode.PERSIST_STREAM
    err_on_unauthorized: bool = False
    enable_auto_commit: bool = False
    auto_index_persist_interval: TimeSpan = 1 * TimeSpan.SECOND


class WriterRequest(Payload):
    config: WriterConfig | None = None
    command: WriterCommand
    frame: FramePayload | None = None


class WriterResponse(Payload):
    command: WriterCommand
    end: TimeStamp | None
    err: ExceptionPayload


class WSWriterCodec(WSFramerCodec):
    def encode(self, pld: Message[WriterRequest]) -> bytes:
        if pld.type == "close" or pld.payload.command != WriterCommand.WRITE:
            data = self.lower_perf_codec.encode(pld)
            return bytes([LOW_PERF_SPECIAL_CHAR]) + data
        data = self.codec.encode(pld.payload.frame, 1)
        data = bytearray(data)
        data[0] = HIGH_PERF_SPECIAL_CHAR
        return bytes(data)

    def decode(self, data: bytes, pld_t: Message[WriterResponse]) -> object:
        if data[0] == LOW_PERF_SPECIAL_CHAR:
            return self.lower_perf_codec.decode(data[1:], pld_t)
        frame = self.codec.decode(data, 1)
        msg = Message[WriterRequest](type="data")
        msg.payload = Payload(command=WriterCommand.WRITE, frame=frame)
        return msg


def parse_writer_mode(mode: CrudeWriterMode) -> WriterMode:
    if mode == "persist_stream":
        return WriterMode.PERSIST_STREAM
    if mode == "persist":
        return WriterMode.PERSIST
    if mode == "stream":
        return WriterMode.STREAM
    if isinstance(mode, WriterMode):
        return mode
    if isinstance(mode, int):
        try:
            return WriterMode(mode)
        except:
            ...
    raise ValueError(f"invalid writer mode {mode}")


ALWAYS_INDEX_PERSIST_ON_AUTO_COMMIT: TimeSpan = TimeSpan(-1)


class WriterClosed(BaseException): ...


class Writer:
    """A writer is used to write telemetry to a set of channels in time order. It
    should not be constructed directly, and should instead be created using the Synnax
    client's open_writer method.

    The writer is a streaming protocol that is heavily optimized for performance. This
    comes at the cost of increased complexity, and should only be used directly when
    writing large volumes of data (such as recording telemetry from a sensor or
    ingesting data from a file). Simpler methods (such as the frame writer's write
    method) should be used in most cases.

    For a detailed guide on writing data to Synnax, see
    https://docs.synnaxlabs.com/reference/concepts/writes. A rough summary of the write
    process is detailed below:

    1. The writer is opened with a starting timestamp and a list of channel keys (or
    names). The writer will fail to open if the starting timestamp overlaps with any
    existing telemetry for any of the channels specified. If the writer is opened
    successfully, the caller is then free to write frames to the writer.

    2. To write a frame, the caller can use the write method and follow the validation
    rules described in the method's documentation. This process is asynchronous,
    meaning that write will return before the frame has been written to the cluster. This
    also means that the writer can accumulate an error after write is called. If the
    writer accumulates an error, all subsequent write and commit calls will return False.
    The caller can check for errors by calling the error method, which returns the
    accumulated error and resets the writer for future use. The caller can also check
    for errors by closing the writer, which will raise any accumulated error.

    3. To commit the written frames to the database, the caller can call the commit
    method. Unlike write, commit is synchronous, meaning that it will not return until
    all frames have been committed to the database. If the writer has accumulated an
    error, commit will return False. After the caller acknowledges the error, they can
    attempt to commit again. Commit can be called several times throughout a writer's
    lifetime, and will commit all frames written since the last commit.

    4. A writer MUST be closed after use in order to prevent resource leaks. Close
    should typically be called in a 'finally' block. If the writer has accumulated an
    error, close will raise the accumulated error.
    """

    _ENDPOINT = "/frame/write"
    _stream: Stream[WriterRequest, WriterResponse]
    _adapter: WriteFrameAdapter
    _close_exc: Exception | None = None

    start: CrudeTimeStamp

    def __init__(
        self,
        start: CrudeTimeStamp,
        client: WebsocketClient,
        adapter: WriteFrameAdapter,
        name: str = "",
        authorities: list[Authority] | Authority = Authority.ABSOLUTE,
        mode: CrudeWriterMode = WriterMode.PERSIST_STREAM,
        err_on_unauthorized: bool = False,
        enable_auto_commit: bool = False,
        auto_index_persist_interval: TimeSpan = 1 * TimeSpan.SECOND,
        use_experimental_codec: bool = True,
    ) -> None:
        self.start = start
        self._adapter = adapter
        if use_experimental_codec:
            client = client.with_codec(WSWriterCodec(adapter.codec))
        self._stream = client.stream(self._ENDPOINT, WriterRequest, WriterResponse)
        config = WriterConfig(
            control_subject=Subject(name=name, key=str(uuid4())),
            keys=self._adapter.keys,
            start=TimeStamp(self.start),
            authorities=normalize(authorities),
            mode=parse_writer_mode(mode),
            err_on_unauthorized=err_on_unauthorized,
            enable_auto_commit=enable_auto_commit,
            auto_index_persist_interval=auto_index_persist_interval,
        )
        exc = self._stream.send(
            WriterRequest(command=WriterCommand.OPEN, config=config)
        )
        if exc is not None:
            raise exc
        _, exc = self._stream.receive()
        if exc is not None:
            raise exc

    @overload
    def write(self, channels_or_data: ChannelName, series: CrudeSeries): ...

    @overload
    def write(
        self, channels_or_data: ChannelKeys | ChannelNames, series: list[CrudeSeries]
    ): ...

    @overload
    def write(
        self,
        channels_or_data: CrudeFrame,
    ): ...

    def write(
        self,
        channels_or_data: (
            ChannelName | ChannelKey | ChannelKeys | ChannelNames | CrudeFrame
        ),
        series: CrudeSeries | list[CrudeSeries] | None = None,
    ) -> None:
        """Writes the given data to the database. The formats are listed below. Before
        we get into them, here are some important terms to know.

            1. Channel ID -> the key or name of the channel(s) you're writing to.
            2. Series or CrudeSeries -> the data for that channel, which can be
            represented as a synnax Series type, a numpy array, or a simple Python
            list. You can also provide a single numeric (or, in the case of variable
            length types, a string or JSON) value and Synnax will convert it into a
            Series for you.

        Here are the formats you can use to write data to a Synnax cluster:

            1. Channel ID and a single series: Writes the series for the given channel.
            2. A list of channel ids and their corresponding series: Assumes a
            one to one mapping of ids to series i.e. the channel id at index i
            corresponds to the series at index i.
            3. A Synnax Frame (see the Frame documentation for more).
            4. A dictionary of channel ids to series i.e. write the series for the
            given channel id.
            5. A pandas DataFrame where the columns are the channel ids and the rows
            are the series to write.
            6. A dictionary of channel ids to a single
            numeric value. Synnax will convert this into a series for you.

        There are a few important rules to keep in mind when writing data to a Synnax
        cluster:

            1. Have exactly one array for each key in the list of keys provided to the
            writer's open method.
            2. Have equal length arrays for each key.
            3. When writing to an index (i.e. TimeStamp) channel, the values must be
            monotonically increasing.

        :returns: False if the writer has accumulated an error. If this is the case,
        the caller should acknowledge the error by calling the error method or closing
        the writer.
        """
        if self._close_exc is not None:
            raise self._close_exc
        frame = self._adapter.adapt(channels_or_data, series)
        try:
            self._exec(
                WriterRequest(command=WriterCommand.WRITE, frame=frame.to_payload()),
                timeout=0,
            )
        except TimeoutError:
            ...

    @overload
    def set_authority(self, value: CrudeAuthority) -> bool: ...

    @overload
    def set_authority(
        self,
        value: ChannelKey | ChannelName,
        authority: CrudeAuthority,
    ) -> bool: ...

    @overload
    def set_authority(
        self,
        value: dict[ChannelKey | ChannelName | ChannelPayload, CrudeAuthority],
    ) -> bool: ...

    def set_authority(
        self,
        value: (
            dict[ChannelKey | ChannelName | ChannelPayload, CrudeAuthority]
            | ChannelKey
            | ChannelName
            | CrudeAuthority
        ),
        authority: CrudeAuthority | None = None,
    ) -> None:
        if self._close_exc is not None:
            raise self._close_exc
        if isinstance(value, int) and authority is None:
            cfg = WriterConfig(keys=[], authorities=[value])
        else:
            if isinstance(value, (ChannelKey, ChannelName)):
                if authority is None:
                    raise ValueError(
                        "authority must be provided when setting a single channel"
                    )
                value = {value: authority}
            value = self._adapter.adapt_dict_keys(value)
            cfg = WriterConfig(
                keys=list(value.keys()),
                authorities=list(value.values()),
            )
        self._exec(WriterRequest(command=WriterCommand.SET_AUTHORITY, config=cfg))

    def commit(self) -> TimeStamp:
        """Commits the written frames to the database. Commit is synchronous, meaning
        that it will not return until all frames have been committed to the database.

        :returns: False if the commit failed due to an error. In this case, the caller
        should acknowledge the error by calling the error method or closing the writer.
        After the error is acknowledged, the caller can attempt to commit again.
        """
        if self._close_exc is not None:
            raise self._close_exc
        res = self._exec(WriterRequest(command=WriterCommand.COMMIT))
        return res.end

    def close(self):
        """Closes the writer, raising any accumulated error encountered during
        operation. A writer MUST be closed after use, and this method should probably
        be placed in a 'finally' block.
        """
        return self._close(None)

    def _close(self, exc: Exception | None) -> None:
        if self._close_exc is not None:
            if isinstance(self._close_exc, WriterClosed):
                return
            raise self._close_exc
        self._close_exc = exc
        self._stream.close_send()
        while True:
            if self._close_exc is not None:
                if isinstance(self._close_exc, WriterClosed):
                    return
                raise self._close_exc
            res, exc = self._stream.receive()
            if exc is not None:
                self._close_exc = WriterClosed() if isinstance(exc, EOF) else exc
            else:
                self._close_exc = decode_exception(res.err)

    def _exec(
        self, req: WriterRequest, timeout: int | None = None
    ) -> WriterResponse | None:
        exc = self._stream.send(req)
        if exc is not None:
            return self._close(exc)
        while True:
            res, exc = self._stream.receive(timeout)
            if exc is not None:
                return self._close(exc)
            exc = decode_exception(res.err)
            if exc is not None:
                return self._close(exc)
            if res.command == req.command:
                return res

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()



#  Copyright 2025 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.
import time
from pathlib import Path
from typing import Any

import click
import numpy as np

from synnax import Synnax
from synnax.channel import Channel
from synnax.cli import default
from synnax.cli.channel import (
    channel_name_table,
    maybe_select_channel,
    prompt_group_channel_names,
)
from synnax.cli.connect import connect_client
from synnax.cli.flow import Context, Flow
from synnax.cli.io import prompt_file
from synnax.cli.telem import select_data_type
from synnax.cli.warning import warning
from synnax.ingest.row import RowIngestionEngine
from synnax.io import IO_FACTORY, ChannelMeta, IOFactory, ReaderType, RowFileReader
from synnax.telem import DataType, Rate, TimeRange, TimeStamp

# A shorthand grouping to represent all values in a set.
GROUP_ALL = "__all__"


@click.command()
@click.pass_context
@click.argument("path_", type=click.Path(exists=True), required=False, default=None)
def ingest(ctx: click.Context, path_: str | None) -> None:
    warning(ctx)
    return pure_ingest(path_)


def pure_ingest(
    path_: Path | str | None,
    client: Synnax | None = None,
    ctx: Context = default.context(),
) -> None:
    flow = Flow(ctx)
    flow.add("initialize_reader", initialize_reader)
    flow.add("connect_client", _connect_client)
    flow.add("ingest_all", ingest_all)
    flow.add("channels_to_ingest", channels_to_ingest)
    flow.add("skip_invalid_channels", skip_invalid_channels)
    flow.add("validate_channels_exist", validate_channels_exist)
    flow.add("validate_data_types", validate_data_types)
    flow.add("validate_start_time", validate_start_time)
    flow.add("create_channels", create_channels)
    flow.add("prompt_name", prompt_name)
    flow.add("ingest", run_ingestion)
    path = None if path_ is None else Path(path_)
    flow.run(IngestionCLI(IO_FACTORY, path, client), "initialize_reader")


class IngestionCLI:
    name: str | None = None
    path: Path | None = None
    factory: IOFactory
    reader: RowFileReader | None
    client: Synnax | None
    filtered_channels: list[ChannelMeta] | None
    not_found: list[ChannelMeta] | None
    db_channels: list[Channel] | None
    start: TimeStamp | None = None

    def __init__(self, factory: IOFactory, path: Path | None, client: Synnax | None):
        self.path = path
        self.factory = factory
        self.reader = None
        self.client = client
        self.filtered_channels = None
        self.not_found = None
        self.db_channels = None


def run_ingestion(ctx: Context, cli: IngestionCLI) -> None:
    """Runs the ingestion process."""
    assert cli.reader is not None
    assert cli.db_channels is not None
    assert cli.client is not None
    assert cli.start is not None
    filtered_reader = initialize_reader(ctx, cli)
    if cli.reader.type() == ReaderType.Row:
        engine = RowIngestionEngine(cli.client, cli.reader, cli.db_channels, cli.start)
    else:
        raise NotImplementedError("Only row ingestion is supported at this time.")
    ctx.console.info("Starting ingestion process...")
    engine.run()
    cli.client.ranges.create(name=cli.name, time_range=TimeRange(cli.start, engine.end))


def initialize_reader(
    ctx: Context,
    cli: IngestionCLI,
) -> str | None:
    ctx.console.info("Welcome to the Synnax ingestion CLI! Let's get started.")
    if cli.path is None:
        ctx.console.info("Please select a file to ingest.")
        cli.path = prompt_file(ctx)
    if cli.path is None:
        return None
    cli.reader = cli.factory.new_reader(cli.path)
    return "connect_client"


def _connect_client(ctx: Context, cli: IngestionCLI) -> str | None:
    """Prompts the user to connect to a Synnax py."""
    if cli.client is None:
        cli.client = connect_client(ctx)
    return "ingest_all" if cli.client else None


def ingest_all(ctx: Context, cli: IngestionCLI) -> str | None:
    """Prompts the user to ingest all channels."""
    assert cli.reader is not None
    if ctx.console.ask("Would you like to ingest all channels?", default=True):
        cli.filtered_channels = cli.reader.channels()
        return "skip_invalid_channels"
    else:
        return "channels_to_ingest"


def channels_to_ingest(ctx: Context, cli: IngestionCLI) -> str | None:
    """Prompts the user to select channels to ingest."""
    assert cli.reader is not None
    ctx.console.info("Which channels would you like to ingest?")
    channels = cli.reader.channels()
    grouped = prompt_group_channel_names(ctx, [ch.name for ch in channels])
    if grouped is None or len(grouped) == 0:
        return None
    all_names = [v for l in grouped.values() for v in l]
    cli.filtered_channels = [ch for ch in channels if ch.name in all_names]
    return "skip_invalid_channels"


def skip_invalid_channels(ctx: Context, cli: IngestionCLI) -> str | None:
    assert cli.reader is not None
    data_types = {
        key: dt
        for key, dt in read_data_types(ctx, cli).items()
        if dt == DataType.UNKNOWN
    }
    if len(data_types) > 0:
        ctx.console.info("The following channels have non-numeric data types")
        channel_name_table(ctx, [ch for ch in data_types.keys()])
        if not ctx.console.ask("Skip these channels?", default=True):
            return None
        cli.filtered_channels = [
            ch for ch in cli.filtered_channels if ch.name not in data_types.keys()
        ]
    return "validate_channels_exist"


def cannot_cast_error(ctx: Context, actual: Any, ch: Channel) -> None:
    ctx.console.error(
        f"""Unable to cast column data type {actual}
for {ch.name} to channel data type {ch.data_type.np}"""
    )


def validate_data_types(ctx: Context, cli: IngestionCLI) -> str | None:
    """Does an optimistic check on the first sample of each channel.  This isn't error
    prone, but checking every sample in a large file would be too slow.
    """
    assert cli.db_channels is not None
    d_types = read_data_types(ctx, cli)

    for ch in cli.db_channels:
        samples_type = d_types[ch.name].np
        ch_type = ch.data_type.np
        if not np.can_cast(samples_type, ch_type):
            return cannot_cast_error(ctx, samples_type, ch)
        elif samples_type != ch_type:
            ctx.console.warn(
                f"""Channel {ch.name} has data type {ch_type} but the file data type is
                {samples_type}. Synnax can cast between them safely, but you have been
                warned."""
            )

    return "validate_start_time"


def read_data_types(ctx: Context, cli: IngestionCLI) -> dict[str, DataType]:
    assert cli.reader is not None
    assert cli.filtered_channels is not None

    cli.reader.set_chunk_size(1)
    cli.reader.seek_first()

    first = cli.reader.read()

    data_types = {}
    for ch in cli.filtered_channels:
        samples = first[ch.name]
        if len(samples) == 0:
            ctx.console.warn(f"Channel {ch.name} has no samples")
            continue
        try:
            data_types[ch.name] = DataType(samples.to_numpy().dtype)
        except TypeError:
            data_types[ch.name] = DataType.UNKNOWN
    return data_types


def validate_channels_exist(ctx: Context, cli: IngestionCLI) -> str | None:
    """Validates that all channels in the file exist in the database. If not, prompts
    the user to create them.
    """
    assert cli.filtered_channels is not None
    assert cli.client is not None

    ctx.console.info("Validating that channels exist...")
    cli.not_found = list()
    cli.db_channels = list()
    for channel in cli.filtered_channels:
        ch = maybe_select_channel(
            ctx,
            cli.client.channels.retrieve([channel.name]),
            channel.name,
        )
        if ch is None:
            cli.not_found.append(channel)
        else:
            cli.db_channels.append(ch)

    if len(cli.not_found) > 0:
        ctx.console.info("The following channels were not found in the database:")
        channel_name_table(ctx, [ch.name for ch in cli.not_found])
        if not ctx.console.ask("Would you like to create them?", default=True):
            return None
        return "create_channels"

    return "validate_data_types"


def validate_start_time(ctx: Context, cli: IngestionCLI) -> str | None:
    """Reads the starting timestamp of the file and prompts the user to confirm that it
    is correct. This is mainly a sanity check to make sure the timestamps are properly
    formatted."""
    assert cli.db_channels is not None
    assert cli.reader is not None
    _idx = [ch for ch in cli.db_channels if ch.is_index]
    if len(_idx) == 0:
        # If there is no index, it means all channels are rate based or we've already
        # written the index data. In either case, we need to prompt the user to enter
        # the start timestamp.
        _start = ctx.console.ask(
            """Please enter the start timestamp of the file as a
            nanosecond UTC integer. If you'd like a converter,
            use https://www.epochconverter.com/""",
            default=TimeStamp.now(),
        )
        cli.start = TimeStamp(_start)
    else:
        idx = _idx[0]
        cli.reader.set_chunk_size(1)
        cli.reader.seek_first()
        first = cli.reader.read()
        cli.start = TimeStamp(first[idx.name].to_numpy()[0])

    ctx.console.info(f"Identified start timestamp for file as {cli.start}.")
    if not ctx.console.ask("Is this correct?", default=True):
        return None
    return "prompt_name"


def create_indexes(
    ctx: Context,
    cli: IngestionCLI,
    options: list[ChannelMeta],
) -> list[ChannelMeta]:
    """Prompts the user to create index channels."""
    assert cli.client is not None
    assert cli.db_channels is not None
    grouped = prompt_group_channel_names(ctx, [ch.name for ch in options])
    if grouped is None:
        return options
    names = [name for v in grouped.values() for name in v]
    for name in names:
        ch = cli.client.channels.create(
            name=name, is_index=True, data_type=DataType.TIMESTAMP
        )
        cli.db_channels.append(ch)
    return [ch for ch in options if ch.name not in names]


DATA_TYPE_OPTIONS = [
    "Guess data types from file",
    "Assign the same data type to all channels (excluding indexes)",
    "Group channels by data type",
]


def assign_data_type(
    ctx: Context,
    cli: IngestionCLI,
) -> dict[DataType, list[ChannelMeta]] | None:
    assert cli.not_found is not None

    grouped = {GROUP_ALL: cli.db_channels}
    assigned = {}
    ctx.console.info("Please select an option for assigning data types:")
    opt, _ = ctx.console.select(
        rows=DATA_TYPE_OPTIONS,
        default=DATA_TYPE_OPTIONS[0],
    )
    if opt == DATA_TYPE_OPTIONS[0]:
        data_types = read_data_types(ctx, cli)
        assigned = {}
        for ch in cli.not_found:
            dt = data_types[ch.name]
            if dt not in assigned:
                assigned[dt] = [ch]
            else:
                assigned[dt].append(ch)
        return assigned
    elif opt == DATA_TYPE_OPTIONS[2]:
        groups = prompt_group_channel_names(ctx, [ch.name for ch in cli.not_found])
        if groups is None or len(groups) == 0:
            return None
        grouped = {
            k: [ch for ch in cli.not_found if ch.name in v] for k, v in groups.items()
        }
    for key, group in grouped.items():
        if key != GROUP_ALL:
            ctx.console.info(f"Assigning data type to {key}")
        dt = select_data_type(ctx)
        assigned[dt] = group
    return assigned


def assign_index_or_rate(
    ctx: Context,
    cli: IngestionCLI,
) -> dict[Rate | str, list[ChannelMeta]] | None:
    """Prompts the user to assign an index/rate to the channels in the given
    group"""
    assert cli.client is not None
    assert cli.not_found is not None
    client = cli.client

    grouped = {GROUP_ALL: cli.not_found}
    if not ctx.console.ask(
        "Do all non-indexed channels have the same data rate or index?",
        bool,
        default=True,
    ):
        if not ctx.console.ask(
            "Can you group channels by data rate or index?", default=True
        ):
            grouped = {v.name: [v] for v in cli.not_found}
        grouped = prompt_group_channel_names(ctx, [ch.name for ch in cli.not_found])
        if grouped is None or len(grouped) == 0:
            return None
        grouped = {
            k: [ch for ch in cli.not_found if ch.name in v] for k, v in grouped.items()
        }

    def assign_to_group(key: str, group: list[ChannelMeta]):
        if key != GROUP_ALL:
            ctx.console.info(f"Assigning data rate or index to {key}")
        _choice = ctx.console.ask("Enter the name of an index or a data rate")
        assert _choice is not None
        if _choice.replace(".", "").isdigit():
            return Rate(float(_choice))
        else:
            # If the user entered a string, we have an index channel, and we
            # need to make sure that the string is a valid index.
            res = client.channels.retrieve([_choice])
            idx = maybe_select_channel(ctx, res, _choice)
            if not idx:
                ctx.console.warn(f"Index channel with key {_choice} not found")
                if ctx.console.ask("Try again?", default=True):
                    return assign_to_group(key, group)
                return None
            return idx.key

    assigned: dict[Rate | str, list[ChannelMeta]] = dict()
    for key, group in grouped.items():
        idx = assign_to_group(key, group)
        if idx is None:
            return None
        assigned[idx] = group

    return assigned


def create_channels(ctx: Context, cli: IngestionCLI) -> str | None:
    assert cli.not_found is not None
    assert cli.client is not None
    assert cli.db_channels is not None

    if ctx.console.ask("Are any channels indexed (e.g. timestamps)?", default=True):
        cli.not_found = create_indexes(ctx, cli, cli.not_found)

    idx_grouped, dt_grouped = assign_index_or_rate(ctx, cli), assign_data_type(ctx, cli)
    if idx_grouped is None or dt_grouped is None:
        return None

    to_create = list()
    for rate_or_index, channels in idx_grouped.items():
        is_rate = isinstance(rate_or_index, Rate)
        for ch in channels:
            to_create.append(
                Channel(
                    name=ch.name,
                    is_index=False,
                    index="" if is_rate else rate_or_index,
                    data_type=[dt for dt, chs in dt_grouped.items() if ch in chs][0],
                )
            )

    cli.db_channels.extend(cli.client.channels.create(to_create))

    return "validate_data_types"


def prompt_name(ctx: Context, cli: IngestionCLI) -> str | None:
    assert cli.db_channels is not None
    assert cli.not_found is not None
    assert cli.client is not None
    path: Path = cli.path
    ctx.console.info("Please enter a name for the data set")
    cli.name = ctx.console.ask("Name", default=path.name)
    return "ingest"
